package cn.com.qmth.stmms.ms.marking.service;

import cn.com.qmth.stmms.ms.core.cache.ParamCache;
import cn.com.qmth.stmms.ms.core.domain.Level;
import cn.com.qmth.stmms.ms.core.domain.task.MarkTaskLevel;
import cn.com.qmth.stmms.ms.core.domain.user.MarkUser;
import cn.com.qmth.stmms.ms.core.repository.LevelRepo;
import cn.com.qmth.stmms.ms.core.repository.MarkUserRepo;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.DoubleStream;
import java.util.stream.Stream;

/**
 * 定档
 * Created by zhengmin on 2016/9/22.
 */
@Service
public class DetermineLevelService {

    public enum DeterType {
        /**
         * 多数
         */
        MAJORITY,
        /**
         * 权重
         */
        WEIGHT
    }

    @Autowired
    private LevelRepo levelRepo;
    @Autowired
    private MarkUserRepo markUserRepo;

    /**
     * 计算定档值（规则：1.不是过半定档，取权重，2.是过半定档，没有取优，取过半定档，3.是过半定档，是取优，取过半定档和权重最优值）
     *
     * @param workId    工作ID
     * @param markTasks 任务集合
     */
    public DeterResult determine(Long workId, List<Level> levels, List<MarkTaskLevel> markTasks) {
        // 是否过半定档（1：是）
        boolean majority = Optional.ofNullable(ParamCache.levelConfigMap.get(String.valueOf(String.valueOf(workId))).getMajority()).orElse(1) == 1;
        DeterResult majorityResult = null;
        DeterResult weightResult = null;
        if (majority) {
            MarkTaskLevel[] tasks = markTasks.toArray(new MarkTaskLevel[0]);
            majorityResult = calcMajority(tasks);
        }
        weightResult = calcWeight(workId, levels, markTasks);

        if (majorityResult != null) {
            // 是否取优（过半定档和权重，取最优值）
            boolean takeBest = Optional.ofNullable(ParamCache.levelConfigMap.get(String.valueOf(String.valueOf(workId))).getTakeBest()).orElse(1) == 1;
            if (takeBest) {
                Map<String, Integer> levelMap = levels.stream().collect(Collectors.toMap(Level::getCode, Level::getLevelValue));
                // 取优
                int majorityLevelValue = levelMap.get(majorityResult.getResult());
                int weightLevelValue = levelMap.get(weightResult.getResult());
                // 值越小，档位越大
                if (majorityLevelValue > weightLevelValue) {
                    return weightResult;
                } else {
                    return majorityResult;
                }
            } else {
                return majorityResult;
            }
        } else {
            return weightResult;
        }
    }

    /**
     * 计算权重值
     *
     * @param markTasks 任务集合
     */
    private DeterResult calcWeight(Long workId, List<Level> levels, List<MarkTaskLevel> markTasks) {
        // 是否开启去高去低再加权评卷（1：是， 0：否，默认0）
        boolean removeHighAndLow = Optional.ofNullable(ParamCache.levelConfigMap.get(String.valueOf(String.valueOf(workId))).getRemoveHighAndLow()).orElse(0) == 1;
        if (removeHighAndLow && markTasks.size() > 2) {
            markTasks.sort((Comparator.comparing(MarkTaskLevel::getResult)));
            markTasks.remove(0);//去掉最大值
            markTasks.remove(markTasks.size() - 1);//去掉最小值
        }

        MarkTaskLevel[] tasks = markTasks.toArray(new MarkTaskLevel[0]);
        //通过权重计算，平均值四舍五入
        double[] values = new double[tasks.length];
        double[] weights = new double[tasks.length];
        for (int i = 0; i < tasks.length; i++) {
            String levelResult = tasks[i].getResult();
            Level level = levels.stream().filter(m -> m.getCode().equals(levelResult)).findFirst().orElse(null);
            if (level == null) {
                throw new RuntimeException("无此档位");
            }
            MarkUser marker = markUserRepo.findOne(tasks[i].getMarkerId());
            double markerWeight = marker.getWeight();
            values[i] = level.getWeight() * markerWeight;
            weights[i] = markerWeight;
        }
        int avg = (int) Math.round(DoubleStream.of(values).sum() / DoubleStream.of(weights).sum());
        Optional<Level> levelOptional = levels.stream().filter(o -> o.getMaxScore() >= avg && o.getMinScore() <= avg).findFirst();
        Level level = levelOptional.orElse(null);
        if (level == null) {
            levels = levels.stream().sorted(Comparator.comparing(Level::getMaxScore).reversed()).collect(Collectors.toList());
            if (avg > levels.get(0).getMaxScore()) {
                level = levels.get(0);
            }
            levels = levels.stream().sorted(Comparator.comparing(Level::getMinScore)).collect(Collectors.toList());
            if (avg < levels.get(0).getMinScore()) {
                level = levels.get(0);
            }
        }
        return new DeterResult(level.getCode(), DeterType.WEIGHT);
    }

    /**
     * 计算过半定档值
     *
     * @param tasks 任务集合
     */
    private DeterResult calcMajority(MarkTaskLevel[] tasks) {
        //如果相同的数量大于半数，直接定档
        Stream<MarkTaskLevel> taskStream = Stream.of(tasks);
        Optional<Map.Entry<String, Long>> optional = taskStream
                .collect(Collectors.groupingBy(MarkTaskLevel::getResult, Collectors.counting()))
                .entrySet()
                .stream()
                .max(Comparator.comparing(Map.Entry::getValue));
        if (optional.isPresent()) {
            Map.Entry<String, Long> entry = optional.get();
            if (entry.getValue() > tasks.length / 2) {
                return new DeterResult(entry.getKey(), DeterType.MAJORITY);
            }
        }
        return null;
    }
}

class DeterResult {
    private String result;
    private DetermineLevelService.DeterType deterType;

    public DeterResult(String result, DetermineLevelService.DeterType deterType) {
        this.result = result;
        this.deterType = deterType;
    }

    public String getResult() {
        return result;
    }

    public void setResult(String result) {
        this.result = result;
    }

    public DetermineLevelService.DeterType getDeterType() {
        return deterType;
    }

    public void setDeterType(DetermineLevelService.DeterType deterType) {
        this.deterType = deterType;
    }
}